param (
    [Parameter(Mandatory=$true)][string]$template,
    [Parameter(Mandatory=$true)][string[]]$md,
    [string]$docx,
    [string]$pdf,
    [switch]$embedfonts,
    [switch]$convertlists
)

if ([string]::IsNullOrEmpty($docx) -and [string]::IsNullOrEmpty($pdf))
{
  Write-error "-docx or -pdf must be specified"
  exit 113
}

$exe = "pandoc.exe"

$template = [System.IO.Path]::GetFullPath($template)

$md = $md | % { [System.IO.Path]::GetFullPath($_ )}

$is_docx_temporary = $False

if ([string]::IsNullOrEmpty($docx))
{
  $docx = [System.IO.Path]::GetTempFilename() + ".docx"
  $is_docx_temporary = $True
}
else
{
  $docx = [System.IO.Path]::GetFullPath($docx)
}

if (-not [string]::IsNullOrEmpty($pdf))
{
  $pdf = [System.IO.Path]::GetFullPath($pdf)
}

$tempdocx = [System.IO.Path]::GetTempFilename() + ".docx"

write-host "Executing Pandoc..."
&$exe $md -o $tempdocx --filter pandoc-crossref --filter pandoc-citeproc --reference-doc $template

if ($LASTEXITCODE -ne 0)
{
  Write-error "Pandoc execution has failed"
  exit 111
}

$word = New-Object -ComObject Word.Application
$curdir = Split-Path $MyInvocation.MyCommand.Path
Set-Location -Path $curdir

$word.ScreenUpdating = $False

$doc = $word.Documents.Open($template)
$doc.Activate()
$selection = $word.Selection

# Save under a new name as soon as possible to prevent auto-saving
# (and polluting) the template
write-host "Saving..."
$doc.SaveAs([ref]$docx)
if (-not $?)
{
  $doc.Close()
  $word.Quit()
  exit 112
}

write-host "Inserting main text..."
if ($selection.Find.Execute("%MAINTEXT%^13", $True, $True, $False, $False, $False, $True, `
       [Microsoft.Office.Interop.Word.wdFindWrap]::wdFindContinue, $False, "",   `
       [Microsoft.Office.Interop.Word.wdReplace]::wdReplaceNone))
{
  $start = $Selection.Range.Start
  $Selection.InsertFile($tempdocx)
  $end = $Selection.Range.End
  $inserted_tables = $doc.Range([ref]$start, [ref]$end).Tables

  # Check if there is anything after the main text
  $selection.WholeStory()
  $totalend = $Selection.Range.End

  # If there is nothing after the main text, remove the extra CR which
  # mystically appears out of nowhere in that case
  if ($end -ge ($totalend - 1))
  {
    $selection.Collapse([Microsoft.Office.Interop.Word.wdCollapseDirection]::wdCollapseEnd)  | out-null
    $selection.MoveLeft([Microsoft.Office.Interop.Word.wdUnits]::wdCharacter, 1, `
                        [Microsoft.Office.Interop.Word.wdMovementType]::wdExtend)  | out-null
    $selection.Delete() | out-null
  }
}

 function ConvertTo-Encoding ([string]$From, [string]$To){
      Begin{
          $encFrom = [System.Text.Encoding]::GetEncoding($from)
          $encTo = [System.Text.Encoding]::GetEncoding($to)
      }
      Process{
          $bytes = $encTo.GetBytes($_)
          $bytes = [System.Text.Encoding]::Convert($encFrom, $encTo, $bytes)
          $encTo.GetString($bytes)
      }
  }
write-host "Searching styles..."
foreach ($style in $doc.Styles)
{ 
  switch ($style.NameLocal | ConvertTo-Encoding "windows-1251" "UTF-8")
  {
    'TableStyleTransactions'          {$TableStyleTransactions = $style; break}
    'TableStyleTransactionsNoHeader'  {$TableStyleTransactionsNoHeader = $style; break}
    'Table Caption'                   {$TableCaption = $style; break}
    #"Абстракт"                        {$AbStyle = $style; break}
    #"Заголовок авторы"                {$ZAv = $style; break}
    #'Заголовок англ'                  {$ZAn = $style; break}
    #"Заголовок авторы англ"           {$ZAvAn = $style; break}
    #"Заголовок литературы"            {$ZL = $style; break}
    "Заголовок подподстатьи"          {$ZPP = $style; break}
    "Заголовок подстатьи"             {$ZP = $style; break}
    #'Заголовок раздела'               {$ZRz = $style; break}
    #"Заголовок работа"                {$ZRb = $style; break}
    #"Заголовок статьи"                {$ZS = $style; break}
    "Заголовок таблицы 1"             {$ZT1 = $style; break}
    "Заголовок таблицы 2"             {$ZT2 = $style; break}
    #'ключевые слова'                  {$KS = $style; break}
    #'Ключевые слова_англ'             {$KSA = $style; break}
    #"Литература"                      {$L = $style; break}
    'Подпись к рисунку'               {$PR = $style; break}
    'Рисунок'                         {$R = $style; break}
    "Текст программы"                 {$TP = $style; break}
    "Текст трудов"                    {$TT = $style; break}
    "Текст трудов список"             {$TTS = $style; break}
    "Текст трудов список_2"           {$TTS2 = $style; break}
    'Формула'                         {$F = $style; break}
    'Формула_номер'                   {$FN = $style; break}
    
    "Основной текст"                  {$Osn = $style; break}
    'First Paragraph'                 {$FP = $style; break}
    'Image Caption'                   {$ImC = $style; break}
    'Captioned Figure'                {$CFig = $style; break}
    "Заголовок 1"                     {$Z1 = $style; break}
    "Заголовок 2"                     {$Z2 = $style; break}
    'Source Code'                     {$SCod = $style; break}
  }
}

$Selection.Find.ClearFormatting | out-null
$Selection.Find.Style = $TableCaption
$Selection.Find.Replacement.ClearFormatting | out-null
$Selection.Find.Replacement.Style = $ZT1
$selection.Find.Execute("^v ", $False, $False, $False, $False, $False, $True, `
       [Microsoft.Office.Interop.Word.wdFindWrap]::wdFindContinue, $True, "^p",   `
       [Microsoft.Office.Interop.Word.wdReplace]::wdReplaceAll) | out-null
  
$selection.Find.Execute("^v ", $False, $False, $False, $False, $False, $True, `
       [Microsoft.Office.Interop.Word.wdFindWrap]::wdFindContinue, $False, ". ",   `
       [Microsoft.Office.Interop.Word.wdReplace]::wdReplaceAll) | out-null

 

$BodyText = [Microsoft.Office.Interop.Word.wdBuiltinStyle]::wdStyleBodyText
$Heading1 = [Microsoft.Office.Interop.Word.wdBuiltinStyle]::wdStyleHeading1
$Heading2 = [Microsoft.Office.Interop.Word.wdBuiltinStyle]::wdStyleHeading2
$Heading3 = [Microsoft.Office.Interop.Word.wdBuiltinStyle]::wdStyleHeading3


function Rename_Style ([System.__ComObject]$From, [System.__ComObject]$To){
          $Selection.Find.ClearFormatting | out-null
          $Selection.Find.Style = $From
          $Selection.Find.Replacement.ClearFormatting | out-null
          $Selection.Find.Replacement.Style = $To
          $selection.Find.Execute("", $False, $False, $False, $False, $False, $True, `
              [Microsoft.Office.Interop.Word.wdFindWrap]::wdFindContinue, $True, "",   `
              [Microsoft.Office.Interop.Word.wdReplace]::wdReplaceAll) | out-null
  }

Rename_Style $Osn $TT | out-null
Rename_Style $FP $TT | out-null
Rename_Style $TableCaption $ZT2 | out-null
Rename_Style $ImC $PR | out-null
Rename_Style $CFig $R | out-null
Rename_Style $Z1 $ZP | out-null
Rename_Style $Z2 $ZPP | out-null
Rename_Style $SCod $TP | out-null

write-host "Handling list templates..."

$bullets = [char]0x2014,[char]0x2014,[char]0x2014,[char]0x2014
$numberposition = 0,0.7,1.4,2.5
$textposition = 0.7,1.4,2.5,3.8
$tabposition = 0.7,1.4,2.5,3.8
$format_nested = "%1)","%1.%2)","%1.%2.%3)","%1.%2.%3.%4)"
$format_headers = "%1","%1.%2","%1.%2.%3","%1.%2.%3.%4"
$format_single = "%1)","%2)","%3)","%4)"

foreach ($tt in $doc.ListTemplates)
{
  for ($il = 1; $il -le $tt.ListLevels.Count -and $il -le 4; $il++)
  {
    $level = $tt.ListLevels.Item($il)
    $bullet = ($level.NumberStyle -eq [Microsoft.Office.Interop.Word.wdListNumberStyle]::wdListNumberStyleBullet)
    $arabic = ($level.NumberStyle -eq [Microsoft.Office.Interop.Word.wdListNumberStyle]::wdListNumberStyleArabic)
    $roman  = ($level.NumberStyle -eq [Microsoft.Office.Interop.Word.wdListNumberStyle]::wdListNumberStyleLowercaseRoman)
    
    if ($bullet)
    {
      if ($level.NumberFormat -ne " ")
      {
        $level.NumberFormat = $bullets[$il - 1] + ""
      }
      $level.NumberPosition = $word.CentimetersToPoints($numberposition[$il - 1])
      $level.Alignment = [Microsoft.Office.Interop.Word.wdListLevelAlignment]::wdListLevelAlignLeft
      $level.TextPosition = $word.CentimetersToPoints($textposition[$il - 1])
      $level.TabPosition = $word.CentimetersToPoints($tabposition[$il - 1])
      $level.ResetOnHigher = $il - 1
      $level.StartAt = 1
      $level.Font.Size = 10
      $level.Font.Name = "PT Serif"
      $level.LinkedStyle = ""
      $level.TrailingCharacter = [Microsoft.Office.Interop.Word.wdTrailingCharacter]::wdTrailingTab
    }
    
    if (($arabic -and ($level.NumberFormat -ne $format_headers[$il - 1])) -or $roman)
    {
      if ($level.NumberFormat -ne " " )
      {
        if ($arabic)
        {
          $level.NumberFormat = $format_nested[$il - 1]
        }
        if ($roman)
        {
          $level.NumberStyle = [Microsoft.Office.Interop.Word.wdListNumberStyle]::wdListNumberStyleArabic;
          $level.NumberFormat = $format_single[$il - 1]
        }
      }
      $level.NumberPosition = $word.CentimetersToPoints($numberposition[$il - 1])
      $level.Alignment = [Microsoft.Office.Interop.Word.wdListLevelAlignment]::wdListLevelAlignLeft
      $level.TextPosition = $word.CentimetersToPoints($textposition[$il - 1])
      $level.TabPosition = $word.CentimetersToPoints($tabposition[$il - 1])
      $level.ResetOnHigher = $il - 1
      $level.StartAt = 1
      $level.Font.Size = 10
      $level.Font.Name = "PT Serif"
      $level.LinkedStyle = ""
      $level.TrailingCharacter = [Microsoft.Office.Interop.Word.wdTrailingCharacter]::wdTrailingTab
    }
  }
}

# Disable grammar checking (it takes time and spews out error messages)
$doc.GrammarChecked = $True

$ntables = 0



write-host "Handling tables..."

# Loop over other tables
for ($t = 1; $t -le $inserted_tables.Count; $t++)
{
  $table = $inserted_tables.Item($t)
  
  # This is to fix the widths of the columns
  $table.AllowAutoFit = $False
	  
  # Numbered equations are 2-column tables without titles and borders, and with equations in both columns
  if ([string]::IsNullOrEmpty($table.Title) `
    -and (-not $table.Rows.Item(1).Cells.Borders.Item([Microsoft.Office.Interop.Word.wdBorderType]::wdBorderBottom).Visible) `
	  -and ($table.Columns.Count -eq 2) `
	  -and ($table.Cell(1,1).Range.OMaths.Count -eq 1) `
	  -and ($table.Cell(1,2).Range.OMaths.Count -eq 1))
  {
    # There can be multiple equations (rows) in one table
    foreach ($row in $table.Rows)
    {
      # After removing the equation, the text contents remains
      if ($row.Cells.Item(2).Range.OMaths.Count -ne 0)
      {
        $row.Cells.Item(2).Range.OMaths.Item(1).Remove()
      }
                
      $row.Cells.Item(2).VerticalAlignment = [Microsoft.Office.Interop.Word.wdCellVerticalAlignment]::wdCellAlignVerticalCenter;
      $row.Cells.Item(2).Select()
      $selection.ClearParagraphAllFormatting()
      $pf = $selection.paragraphFormat
      $pf.LeftIndent = $word.CentimetersToPoints(0)
      $pf.RightIndent = $word.CentimetersToPoints(0)
      $pf.Alignment = [Microsoft.Office.Interop.Word.wdParagraphAlignment]::wdAlignParagraphRight
      $pf.SpaceBefore = 0
      $pf.SpaceBeforeAuto = $False
      $pf.SpaceAfter = 0
      $pf.SpaceAfterAuto = $False
      $pf.LineSpacingRule = [Microsoft.Office.Interop.Word.wdLineSpacing]::wdLineSpaceSingle
      $pf.CharacterUnitLeftIndent = 0
      $pf.CharacterUnitRightIndent = 0
      $pf.LineUnitBefore = 0
      $pf.LineUnitAfter = 0
    }
  }
  else # Ordinary tables
  {
    # Count only tables with title (and number)
    if (-not [string]::IsNullOrEmpty($table.Title))
    {
      $ntables = $ntables + 1
    }
    
    $table.Select()
    
    $pf = $selection.paragraphFormat
    $pf.LineSpacingRule = [Microsoft.Office.Interop.Word.wdLineSpacing]::wdLineSpaceSingle
    
    # If the first row has line under it, then it is a table with a header row
    if ($table.Rows.Item(1).Cells.Borders.Item([Microsoft.Office.Interop.Word.wdBorderType]::wdBorderBottom).Visible)
    {
      $table.Style = $TableStyleTransactions
    }
    else # table without header row
    {
      $table.Style = $TableStyleTransactionsNoHeader
    }
    
    # Fix alignment of display-math objects so they math the aligment of text in the cells
    foreach ($row in $table.Rows)
    {
      foreach ($cell in $row.Cells)
      {
        foreach ($math in $cell.Range.OMaths)
        {
          if ($math.Type -eq [Microsoft.Office.Interop.Word.wdOMathType]::wdOMathDisplay)
          {
            $al = $cell.Range.ParagraphFormat.Alignment
            if ($al -eq [Microsoft.Office.Interop.Word.wdParagraphAlignment]::wdAlignParagraphRight)
            {
              $math.Justification = [Microsoft.Office.Interop.Word.wdOMathJc]::wdOMathJcRight;
            }
            elseif ($al -eq [Microsoft.Office.Interop.Word.wdParagraphAlignment]::wdAlignParagraphLeft)
            {
              $math.Justification = [Microsoft.Office.Interop.Word.wdOMathJc]::wdOMathJcLeft;
            }
            elseif ($al -eq [Microsoft.Office.Interop.Word.wdParagraphAlignment]::wdAlignParagraphCenter)
            {
              $math.Justification = [Microsoft.Office.Interop.Word.wdOMathJc]::wdOMathJcCenter;
            }
          }
        }
      }
    }
  }
}


write-host "Updating paragraph styles..."
foreach ($par in $doc.Paragraphs)
{
  $namelocal = $par.Style.NameLocal | ConvertTo-Encoding "windows-1251" "UTF-8"
  # Make source core paragraphs smaller
  if ($namelocal -eq "Текст программы")
  {
    $par.Range.Font.Size = 8
  }
  if ($namelocal -eq "Литература")
  {
    $par.Range.Font.Size = 9
  }
}


write-host "Increasing math font size..."

foreach ($math in $doc.OMaths)
{
  # Size equations up a bit to match Paratype font size
  $math.Range.Font.Size = 10.5
}

# Convert lists numers to plain text (to facilitate merging multiple articles into one)
if ($convertlists)
{
  $selection.WholeStory()
  $selection.Range.ListFormat.ConvertNumbersToText()
}

if ($embedfonts)
{
  # Embed fonts (for users who do not have Paratype fonts installed).
  # This costs a few MB in file size
  $word.ActiveDocument.EmbedTrueTypeFonts = $True
  $word.ActiveDocument.DoNotEmbedSystemFonts = $True
  $word.ActiveDocument.SaveSubsetFonts = $True 
}

if (-not $is_docx_temporary)
{
  write-host "Saving docx..."
  $doc.Save()
}

if (-not [string]::IsNullOrEmpty($pdf))
{
  write-host "Saving PDF..."
  $doc.SaveAs2([ref]$pdf, [ref][Microsoft.Office.Interop.Word.wdSaveFormat]::wdFormatPDF)
}

$doc.Close()
$word.Quit()

write-host "Removing temporary files..."
Remove-item -path $tempdocx
if ($is_docx_temporary)
{
  Remove-item -path $docx
}
